/*@

fixpoint bool forall<t>(list<t> xs, fixpoint(t, bool) p) {
    switch (xs) {
        case nil: return true;
        case cons(x, xs0): return p(x) && forall(xs0, p);
    }
}

fixpoint bool le(unit u, int x, int y) {
    switch (u) {
        case unit: return x <= y;
    }
}

fixpoint a fold_left<a, b>(a x0, fixpoint(a, b, a) f, list<b> xs) {
    switch (xs) {
        case nil: return x0;
        case cons(x, xs0): return fold_left(f(x0, x), f, xs0);
    }
}

lemma void fold_left_append<a, b>(a x0, fixpoint(a, b, a) f, list<b> xs1, list<b> xs2)
    requires true;
    ensures fold_left(x0, f, append(xs1, xs2)) == fold_left(fold_left(x0, f, xs1), f, xs2);
{
    switch (xs1) {
        case nil:
        case cons(x, xs10):
            fold_left_append(f(x0, x), f, xs10, xs2);
    }
}

fixpoint int max_func(unit u, int x, int y) {
    switch (u) {
        case unit: return x < y ? y : x;
    }
}

fixpoint int plus(unit u, int x, int y) {
    switch (u) {
        case unit: return x + y;
    }
}

predicate ints(int *a, int N; list<int> elems) =
    N == 0 ?
        elems == nil
    :
        0 < N &*&
        integer(a, ?elem) &*&
        ints(a + 1, N - 1, ?elems0) &*&
        elems == cons(elem, elems0);

lemma void ints_split(int i)
    requires ints(?a, ?N, ?elems) &*& 0 <= i &*& i <= N;
    ensures ints(a, i, take(i, elems)) &*& ints(a + i, N - i, drop(i, elems)) &*& elems == append(take(i, elems), drop(i, elems));
{
    if (i == 0) {
        close ints(a, 0, nil);
    } else {
        open ints(a, N, elems);
        ints_split(i - 1);
    }
}

lemma void ints_merge(int *a)
    requires ints(a, ?i, ?elems1) &*& ints(a + i, ?n, ?elems2) &*& 0 <= n;
    ensures ints(a, i + n, append(elems1, elems2));
{
    open ints(a, i, elems1);
    if (i == 0) {
    } else {
        ints_merge(a + 1);
        close ints(a, i + n, append(elems1, elems2));
    }
}

lemma_auto void ints_inv()
    requires ints(?a, ?N, ?elems);
    ensures ints(a, N, elems) &*& 0 <= N &*& N == length(elems);
{
    open ints(a, N, elems);
    if (N != 0) {
        ints_inv();
    }
}

lemma void take_drop_plus_1<t>(int i, list<t> xs)
    requires 0 <= i &*& i < length(xs);
    ensures take(i + 1, xs) == append(take(i, xs), cons(head(drop(i, xs)), nil));
{
    switch (xs) {
        case nil:
        case cons(x, xs0):
            if (i == 0) {
            } else {
                take_drop_plus_1(i - 1, xs0);
            }
    }
}

lemma void le_mult_compat(int x, int y1, int y2);  // NOT PROVEN
    requires 0 <= x &*& y1 <= y2;
    ensures x * y1 <= x * y2;

lemma void mult_congr_l(int x1, int x2, int y);  // NOT PROVEN
    requires x1 == x2;
    ensures x1 * y == x2 * y;

@*/

void problem1(int *a, int N)
    //@ requires ints(a, N, ?elems) &*& forall(elems, (le)(unit, 0)) == true;
    //@ ensures ints(a, N, elems);
{
    int sum = 0;
    int max = 0;
    int i = 0;
    for (; i < N; i++)
        /*@
        invariant
            ints(a, N, elems) &*&
            0 <= i &*& i <= N &*&
            0 <= max &*&
            sum == fold_left(0, (plus)(unit), take(i, elems)) &*&
            max == fold_left(0, (max_func)(unit), take(i, elems)) &*&
            sum <= i * max;
        @*/
    {
        //@ ints_split(i);
        //@ open ints(a + i, _, _);
        //@ assert sum <= i * max;
        int x = *(a + i);
        int oldMax = max;
        if (max < x) {
            max = x;
        }
        //@ assert oldMax <= max;
        //@ assert 0 <= i;
        //@ le_mult_compat(i, oldMax, max);
        //@ assert i * oldMax <= i * max;
        //@ assert sum <= i * max;
        sum += *(a + i);
        //@ close ints(a + i, N - i, _);
        //@ ints_merge(a);
        //@ take_drop_plus_1(i, elems);
        //@ fold_left_append(0, (plus)(unit), take(i, elems), cons(x, nil));
        //@ fold_left_append(0, (max_func)(unit), take(i, elems), cons(x, nil));
    }
    //@ assert i == N;
    //@ mult_congr_l(i, N, max);
    //@ assert i * max == N * max;
    //@ assert sum <= i * max;
    //@ assert sum <= N * max;
}

// Test case: sum is 45.
