method M(N: int, A: array<int>, B: array<int>)
  requires 0 <= N && A != null && B != null && N == |A| && N == |B| && A != B;
  requires (forall k :: 0 <= k && k < N ==> 0 <= A[k] && A[k] < N);
  requires (forall j,k :: 0 <= j && j < k && k < N ==> A[j] != A[k]);  // A is injective
  requires (forall m :: 0 <= m && m < N && inImage(m) ==> (exists k :: 0 <= k && k < N && A[k] == m));  // A is surjective
  modifies B;
  ensures (forall k :: 0 <= k && k < N ==> 0 <= B[k] && B[k] < N);
  ensures (forall k :: 0 <= k && k < N ==> B[A[k]] == k && A[B[k]] == k);  // A and B are each other's inverses
  ensures (forall j,k :: 0 <= j && j < k && k < N ==> B[j] != B[k]);  // (which means that) B is injective
{
  var i := 0;
  while (i < N)
    invariant i <= N;
    invariant (forall k :: 0 <= k && k < i ==> B[A[k]] == k);
  {
    B[A[i]] := i;
    i := i + 1;
  }
  call Lemma(N, A, B);
}

static function inImage(i: int): bool { true }  // this function is used to trigger the surjective quantification

ghost method Lemma(N: int, A: array<int>, B: array<int>)
  requires 0 <= N && A != null && B != null && N == |A| && N == |B|;
  requires (forall k :: 0 <= k && k < N ==> 0 <= A[k] && A[k] < N);
  requires (forall m :: 0 <= m && m < N && inImage(m) ==> (exists k :: 0 <= k && k < N && A[k] == m));  // A is surjective
  requires (forall k :: 0 <= k && k < N ==> B[A[k]] == k);
  ensures (forall j :: 0 <= j && j < N ==> 0 <= B[j] && B[j] < N && A[B[j]] == j);
{
  var n := 0;
  while (n < N)
    invariant n <= N;
    invariant (forall j :: 0 <= j && j < n ==> 0 <= B[j] && B[j] < N && A[B[j]] == j);
  {
    assert inImage(n);
    n := n + 1;
  }
}

method Main()
{
  var a := new int[10];
  a[0] := 9;
  a[1] := 3;
  a[2] := 8;
  a[3] := 2;
  a[4] := 7;
  a[5] := 4;
  a[6] := 0;
  a[7] := 1;
  a[8] := 5;
  a[9] := 6;
  var b := new int[10];
  call M(10, a, b);
  print "a:\n";
  call PrintArray(a);
  print "b:\n";
  call PrintArray(b);
}

method PrintArray(a: array<int>)
  requires a != null;
{
  var i := 0;
  while (i < |a|) {
    print a[i], "\n";
    i := i + 1;
  }
}
