method Search(N: int) returns (success: bool, board: seq<int>)
  requires 0 <= N;
  ensures success ==>
              |board| == N &&
              (forall p :: 0 <= p && p < N ==> IsConsistent(board, p));
  ensures !success ==>
              (forall B ::
                  |B| == N && (forall i :: 0 <= i && i < N ==> 0 <= B[i] && B[i] < N)
                  ==>
                  (exists p :: 0 <= p && p < N && !IsConsistent(B, p)));
{
  call success, board := SearchAux(N, []);
}

static function IsConsistent(board: seq<int>, pos: int): bool
{
  0 <= pos && pos < |board| &&
  (forall q :: 0 <= q && q < pos ==>
      board[q] != board[pos] &&
      board[q] - board[pos] != pos - q &&
      board[pos] - board[q] != pos - q)
}

ghost method Lemma0(short: seq<int>, long: seq<int>, pos: int)
  requires short <= long;
  ensures IsConsistent(short, pos) ==> IsConsistent(long, pos);
{
  if (IsConsistent(short, pos)) {
    assert IsConsistent(long, pos);
  }
}

ghost method Lemma1(short: seq<int>, long: seq<int>)
  requires short <= long;
  ensures (forall k :: 0 <= k && k < |short| && IsConsistent(short, k) ==> IsConsistent(long, k));
  ensures (forall B, p :: IsConsistent(B, p) <==> old(IsConsistent(B, p)));
{
  var j := 0;
  while (j < |short|)
    invariant j <= |short|;
    invariant (forall k :: 0 <= k && k < j && IsConsistent(short, k) ==> IsConsistent(long, k));
  {
    call Lemma0(short, long, j);
    j := j + 1;
  }
}

method CheckConsistent(board: seq<int>, pos: int) returns (b: bool)
  ensures b <==> IsConsistent(board, pos);
  ensures (forall B, p :: IsConsistent(B, p) <==> old(IsConsistent(B, p)));
{
  // stupid as it is, the compiler won't compile a function with a 'forall' inside, so we
  // have to provide a method that computes the same thing as the function defines
  if (0 <= pos && pos < |board|) {
    var p := 0;
    while (p < pos)
      invariant p <= pos;
      invariant (forall q :: 0 <= q && q < p ==>
                  board[q] != board[pos] &&
                  board[q] - board[pos] != pos - q &&
                  board[pos] - board[q] != pos - q);
    {
      if (!(board[p] != board[pos] &&
            board[p] - board[pos] != pos - p &&
            board[pos] - board[p] != pos - p)) {
        b := false;
        return;
      }
      p := p + 1;
    }
    b := true;
    return;
  }
  b := false;
}

method SearchAux(N: int, boardSoFar: seq<int>) returns (success: bool, newBoard: seq<int>)
  requires 0 <= N && |boardSoFar| <= N;
  // consistent so far:
  requires (forall k :: 0 <= k && k < |boardSoFar| ==> IsConsistent(boardSoFar, k));
  ensures success ==>
              |newBoard| == N &&
              (forall p :: 0 <= p && p < N ==> IsConsistent(newBoard, p));
  ensures !success ==>
              (forall B ::
                  |B| == N && (forall i :: 0 <= i && i < N ==> 0 <= B[i] && B[i] < N) &&
                  boardSoFar <= B
                  ==>
                  (exists p :: 0 <= p && p < N && !IsConsistent(B, p)));
  ensures (forall B, p :: IsConsistent(B, p) <==> old(IsConsistent(B, p)));
  decreases N - |boardSoFar|;
{
  var pos := |boardSoFar|;
  if (pos == N) {
    newBoard := boardSoFar;
    success := true;
  } else {
    var n := 0;
    while (n < N)
      invariant n <= N;
      invariant (forall B ::
                  |B| == N && (forall i :: 0 <= i && i < N ==> 0 <= B[i] && B[i] < N) &&
                  boardSoFar <= B &&
                  0 <= B[|boardSoFar|] && B[|boardSoFar|] < n
                  ==>
                  (exists p :: 0 <= p && p < N && !IsConsistent(B, p)));
    {
      var candidateBoard := boardSoFar + [n];
      call consistent := CheckConsistent(candidateBoard, pos);
      if (consistent) {
        call Lemma1(boardSoFar, candidateBoard);
        call s, b := SearchAux(N, candidateBoard);
        if (s) {
          newBoard := b;
          success := true;
          return;
        }
      } else {
        // the following is an unproven lemma:
        assume (forall B ::
                  |B| == N && (forall i :: 0 <= i && i < N ==> 0 <= B[i] && B[i] < N) &&
                  candidateBoard <= B
                  ==>
                  (exists p :: 0 <= p && p < N && !IsConsistent(B, p)));
      }
      n := n + 1;
    }

    success := false;
  }
}

method Main()
{
  call s, b := Search(2);
  print "N=2 returns ", s, "\n";
  call s, b := Search(4);
  print "N=4 returns ", s, "\n";
  call PrintSeq(b);
}

method PrintSeq(b: seq<int>)
{
  var i := 0;
  while (i < |b|) {
    print "  ", b[i], "\n";
    i := i + 1;
  }
}
